/*
 * create multiple gthreads and alarm from each thread
 * return from each thread that catches a signal
 * g_error when an alarm is caught from a different thread than it was set from
 */
#include <glib.h>
#include <signal.h>
#include "helper.c"

#define MYLOOPSIZE 100000
#define NUMTHREADS 20

/* global mutex to serialize thread creation from the main function */
GMutex *_create_mutex = NULL;
GCond *_create_cond = NULL;

/* global hash to map GThread * to our thread numbers */
GHashTable *_thread_table;
GMutex *_thread_table_mutex = NULL;

/* global mutex for alarm raising;
 * set by a thread setting the alarm, and cleared by the alarm signal handler */
GMutex *_alarm_mutex = NULL;

/* last GThread that set the alarm */
GThread *_last_thread_alarm;

/* global variable keeping track of running count of threads */
gint _threads;

/* a signal handler for alarm */
static void
sig_alrm (int signo)
{
  GThread *me;
  const gchar *thread_string;

  me = g_thread_self ();
  thread_string = g_hash_table_lookup (_thread_table, me);
  g_print ("\nCaught alarm from thread %s (%p)\n", thread_string, me);
  /* mutex for alarm is still locked, so the last thread is still valid */
  if (me != _last_thread_alarm)
    g_error ("Thread caught from thread pointer %p while it was set from %p",
             me, _last_thread_alarm);
  g_mutex_unlock (_alarm_mutex);
}

/* endless loop printing thread number and trying to lock the alarm mutex */
static void
endless_print_with_mutex (const char *string)
{
  int i;
  while (1)
  {
    for (i = 0; i < MYLOOPSIZE; ++i)
    {
      if (g_mutex_trylock (_alarm_mutex))
      {
        GThread *me;
        me = g_thread_self ();
        /* only set alarm if we're not the same one as the last to set it
         * so we spread out alarm sets across threads */
        if (_last_thread_alarm == me)
        {
          /* return (ie stop thread) if we're not the main thread */
          if (string[0] != '0')
          {
            --_threads;
            g_print ("Returning from thread %s (%p), %d left\n", string, me, _threads);
            /* we're done forever, so we release the lock */
            g_mutex_unlock (_alarm_mutex);
            return;
          }
          /* return if there are no created threads left */
          if (_threads == 1)
          {
            g_print ("Only main thread left, stopping.\n");
            return;
          }
          /* we don't want to raise the signal, so we release the lock */
          g_mutex_unlock (_alarm_mutex);
        }
        else
        {
          /* set us up a new one */
          g_print ("Setting alarm from thread %s (%p)\n", string, me);
          _last_thread_alarm = me;
          alarm (1);
        }
      }
    }
    g_print (string);
  }
}

/* foreach print_thread function on the hash */
static void
print_thread (gpointer *key, gchar *thread_string)
{
  g_print ("hash_table: thread %s (%p)\n", thread_string, key);
}

/* a thread we're going to create */
gpointer
a_thread (gpointer pointer)
{
  GThread *me;
  const gchar *string;
  gint *numberp;
  gint number;

  g_mutex_lock (_create_mutex);
  /* store all the info we need to keep locally */
  numberp = pointer;
  number = (gint) *numberp;

  /* store the thread number */
  me = g_thread_self ();

  /* in theory we don't need to lock here since we already guarantee
   * serial thread creation */
  g_mutex_lock (_thread_table_mutex);
  g_hash_table_insert (_thread_table, me, g_strdup_printf ("%d", number));
  g_mutex_unlock (_thread_table_mutex);

  g_print ("a_thread: I am thread %d (%p) and I am alive.\n", number, me);

  /* unlock the global mutex so the main thread can go on creating */
  g_cond_signal (_create_cond);
  g_mutex_unlock (_create_mutex);

  g_print ("a_thread: %d (%p): going endless.\n", number, me);
  string = g_strdup_printf ("%d", number);
  endless_print_with_mutex (string);
}
int
main (int argc, char *argv[])
{
  GThread *thread, *me;

  g_thread_init (NULL);

  /* create mutex/cond pair for serialized thread creation */
  _create_mutex = g_mutex_new ();
  _create_cond = g_cond_new ();

  /* create global thread hash and mutex */
  _thread_table = g_hash_table_new (g_direct_hash, g_direct_equal);
  _thread_table_mutex = g_mutex_new ();

  /* create alarm mutex and lock it so nobody starts setting up alarms until
     main fires the first one */
  _alarm_mutex = g_mutex_new ();
  g_mutex_lock (_alarm_mutex);
  

  me = g_thread_self ();
  g_print ("main:     my gthread: %p\n", me);
  /* insert into hash table; no need to lock yet */
  g_hash_table_insert (_thread_table, me, g_strdup ("0"));

  if (signal (SIGALRM, sig_alrm) == SIG_ERR)
    g_error ("Could not set up SIGALRM handler");
  g_print ("main:     I am the main thread and I am alive.\n");

  /* create all threads */
  for (_threads = 1; _threads < NUMTHREADS; ++_threads)
  {
    g_print ("main:     creating thread\n");
    thread = g_thread_create ((GThreadFunc) a_thread, &_threads, TRUE, NULL);
    g_mutex_lock (_create_mutex);
    g_cond_wait (_create_cond, _create_mutex);
    g_mutex_unlock (_create_mutex);
    g_print ("main:     created thread %d: %p\n", _threads, thread);
  }

  /* Dump the hash table of threads */
  g_hash_table_foreach (_thread_table, (GHFunc) print_thread, NULL);

  /* Set the first alarm - mutex was already set at the start */
  g_print ("main:     setting first alarm.\n");
  _last_thread_alarm = me;
  alarm (1);

  g_print ("main:     going endless.\n");
  endless_print_with_mutex ("0");
}
